# type: ignore
"""A test module.

Contains some functions for testing the parser.
"""


def my_func() -> None:
    """The function title.

    This is a function where the ending quotes
    are in the same line as the title"""  # noqa



"""Line break function."""

from typing import List

from docstripy.lines_routines import (
    add_eol,
    clean_trailing_empty,
    clean_trailing_spaces,
)


def line_break(
    lines: List[str], max_line_length: int, num_add_char: int = 0
) -> List[str]:
    """Break lines at a given length.

    Parameters
    ----------
    lines : List[str]
        List of lines to break.
    max_line_length : int
        Maximum line length.
    num_add_char : int, optional
        Number of additional characters to add at the first line that are not
        on the `lines` list. By default, 0.
    """
    if max_line_length <= 0 or not lines:
        return lines
    new_lines = lines.copy()
    new_lines = clean_trailing_spaces(new_lines)
    new_lines[0] = num_add_char * " " + new_lines[0]  # Will be removed at the end
    flat_line = "".join(new_lines)
    list_flat = []
    for i, char in enumerate(flat_line):
        if char != "\n":
            list_flat.append(char)
        else:
            if i > 0 and flat_line[i - 1] == "\n":
                list_flat.append("\n")
            elif i < len(flat_line) - 1:
                next_chr_i = ord(flat_line[i + 1])
                next_is_alpha = 65 <= next_chr_i <= 90 or 97 <= next_chr_i <= 122
                if next_is_alpha:
                    list_flat.append(" ")
                else:
                    list_flat.append("\n")
            else:
                list_flat.append("\n")
    new_lines = "".join(list_flat).split("\n")
    sentences = [line.split(". ") for line in new_lines]
    for paragraph in sentences:
        for i, line_par in enumerate(paragraph):
            line_par += "." if i < len(paragraph) - 1 else ""
            paragraph[i] = line_par
    sentences = [break_sentences(sentence, max_line_length) for sentence in sentences]
    sentences = [add_eol(sentence) for sentence in sentences]
    new_lines = [line for sublist in sentences for line in sublist]
    # Remove leading padding symbol
    new_lines[0] = new_lines[0][num_add_char:]
    new_lines = clean_trailing_empty(new_lines)
    return new_lines


def break_sentences(sentences: List[str], max_line_length: int) -> List[str]:
    """Break sentences at a given length."""
    if not sentences:
        return sentences
    if len(sentences[0]) <= max_line_length:
        if len(sentences[0]) >= 0.75 * max_line_length or len(sentences) == 1:
            # We avoid breaking few words after a dot
            return [sentences[0]] + break_sentences(sentences[1:], max_line_length)
        # We can continue the line with the next sentence
        sentences[1] = sentences[0] + " " + sentences[1]
        return break_sentences(sentences[1:], max_line_length)
    word_split = sentences[0].split(" ")
    line = word_split[0]
    i_word = 1
    while len(line) + len(word_split[i_word]) + 1 <= max_line_length:
        line += " " + word_split[i_word]
        i_word += 1
    remaining_line = " ".join(word_split[i_word:])
    return [line] + break_sentences([remaining_line] + sentences[1:], max_line_length)



"""Docstring building functions for numpy style."""

from typing import Any, Dict, List

from docstripy.line_break import line_break
from docstripy.lines_routines import (
    add_indent,
    clean_trailing_empty,
    clean_trailing_spaces,
)


def build_doc_google(
    current_docstring: List[str],
    sections_dict: Dict[str, Any],
    max_len: int,
    indent: int,
) -> List[str]:
    """Build docstring for google style."""
    docstring = current_docstring.copy()
    section_to_header = {
        "_parameters": "Args:",
        "_raises": "Raises:",
        "_returns": "Returns:",
        "_attributes": "Attributes:",
        "_yields": "Yields:",
    }
    for section_name in (
        "_parameters",
        "_raises",
        "_returns",
        "_yields",
        "_attributes",
    ):
        if section_name in sections_dict and sections_dict[section_name]:
            docstr_sec = []
            docstr_sec.append("\n")
            header = section_to_header[section_name]
            docstr_sec.append(header + "\n")
            doc_params = build_section_params_google(
                sections_dict[section_name],
                max_len=max_len,
                indent=indent,
            )
            if clean_trailing_empty(doc_params):
                docstr_sec.extend(doc_params)
                docstring.extend(docstr_sec)
            else:
                # Error if empty sections
                raise ValueError(f"Empty section found (section {section_name[1:]}).")
    for section_name in sections_dict:
        if not section_name.startswith("_"):
            docstring.append("\n")
            docstring.append(section_name + ":\n")
            section_content = line_break(
                sections_dict[section_name],
                max_line_length=max_len - indent,
            )
            docstring.extend(add_indent(section_content, indent))
    docstring.append('"""\n')
    docstring = clean_trailing_spaces(docstring)
    return docstring


def build_section_params_google(
    param_dicts: List[dict],
    max_len: int,
    indent: int,
) -> List[str]:
    """Build parameters, returns, raises, and attributes sections for google style."""
    docstring = []
    for param_dict in param_dicts:
        param_docstring = []
        first_line = ""
        if "name" in param_dict and param_dict["name"]:
            first_line += param_dict["name"]
        if "type" in param_dict and param_dict["type"]:
            if "name" in param_dict and param_dict["name"]:
                first_line += " ("
            first_line += param_dict["type"]
            if "optional" in param_dict and param_dict["optional"]:
                first_line += ", optional"
            if "name" in param_dict and param_dict["name"]:
                first_line += ")"
        if "description" in param_dict and param_dict["description"]:
            first_line += ": " + param_dict["description"][0].rstrip("\n")
        elif "default" in param_dict and param_dict["default"]:
            first_line += ":"
        param_docstring.append(first_line + "\n")
        if "description" in param_dict and len(param_dict["description"]) > 1:
            param_docstring.extend(param_dict["description"][1:])
        if "default" in param_dict and param_dict["default"]:
            if "description" not in param_dict or not param_dict["description"]:
                line = "Defaults to " + param_dict["default"] + ".\n"
                param_docstring.append(line)
            else:
                param_docstring[-1] = param_docstring[-1].rstrip("\n")
                param_docstring[-1] += " Defaults to " + param_dict["default"] + ".\n"
        param_docstring = line_break(param_docstring, max_len - 2 * indent)
        for i, line in enumerate(param_docstring):
            if i > 0 and line.strip():
                param_docstring[i] = " " * indent + line
        # Global indentation
        param_docstring = [" " * indent + line for line in param_docstring]
        docstring.extend(param_docstring)
    return docstring



"""Docstring building functions for numpy style."""

from typing import Any, Dict, List

from docstripy.line_break import line_break
from docstripy.lines_routines import (
    add_indent,
    clean_trailing_empty,
    clean_trailing_spaces,
)


def build_doc_rest(
    current_docstring: List[str],
    sections_dict: Dict[str, Any],
    max_len: int,
    indent: int,
) -> List[str]:
    """Build docstring for rest style."""
    docstring = current_docstring.copy()
    for section_name in (
        "_parameters",
        "_raises",
        "_returns",
        "_yields",
        "_attributes",
    ):
        if section_name in sections_dict and sections_dict[section_name]:
            doc_params = build_section_params_rest(
                sections_dict[section_name],
                max_len=max_len,
                indent=indent,
                section_name=section_name,
            )
            if clean_trailing_empty(doc_params):
                docstring.extend(doc_params)
            else:
                # Error if empty sections
                raise ValueError(f"Empty section found (section {section_name[1:]}).")
    for section_name in sections_dict:
        if not section_name.startswith("_"):
            docstring.append("\n")
            docstring.append(section_name + ":\n")
            section_content = line_break(
                sections_dict[section_name],
                max_line_length=max_len - indent,
            )
            docstring.extend(add_indent(section_content, indent))
    docstring.append('"""\n')
    docstring = clean_trailing_spaces(docstring)
    return docstring


def build_section_params_rest(
    param_dicts: List[dict],
    max_len: int,
    indent: int,
    section_name: str,
) -> List[str]:
    """Build parameters, returns, raises, and attributes sections for google style."""
    docstring = []
    name_to_keyword = {
        "_parameters": ("param", "type"),
        "_raises": ("raises", ""),
        "_returns": ("return", "rtype"),
        "_attributes": ("atribute", "type"),
        "_yields": ("yield", "rtype"),
    }
    keyword, type_keyword = name_to_keyword[section_name]
    for param_dict in param_dicts:
        param_docstring = []
        first_line = ""
        if ("name" in param_dict and param_dict["name"]) or (
            "description" in param_dict and param_dict["description"]
        ):
            if "name" in param_dict and param_dict["name"]:
                first_line = f":{keyword} {param_dict['name']}:"
            else:
                first_line = f":{keyword}:"
        if "description" in param_dict and param_dict["description"]:
            first_line += " " + param_dict["description"][0].rstrip("\n")
        if first_line:
            param_docstring.append(first_line + "\n")
        if "description" in param_dict and len(param_dict["description"]) > 1:
            param_docstring.extend(param_dict["description"][1:])
        if "default" in param_dict and param_dict["default"]:
            if "description" not in param_dict:
                line = "Defaults to " + param_dict["default"] + ".\n"
                param_docstring.append(line)
            else:
                param_docstring[-1] = param_docstring[-1].rstrip("\n")
                param_docstring[-1] = param_docstring[-1].rstrip(".")
                param_docstring[-1] += ", defaults to " + param_dict["default"] + ".\n"
        param_docstring = line_break(param_docstring, max_len - indent)
        for i, line in enumerate(param_docstring):
            if not line.startswith(":") and line.strip():
                param_docstring[i] = " " * indent + line
        if "type" in param_dict and param_dict["type"]:
            if "name" in param_dict and param_dict["name"]:
                line = f":{type_keyword} {param_dict['name']}:"
            else:
                line = f":{type_keyword}:"
            line += " " + param_dict["type"]
            if "optional" in param_dict and param_dict["optional"]:
                line += ", optional"
            param_docstring.append(line + "\n")
        docstring.extend(param_docstring)
    return docstring



# type: ignore
"""A test module.

Contains some functions for testing the parser.
"""

from typing import Dict, List


def myfunc_rest(  # pylint: disable=W0102
    name1: List[bool],
    name2: list | str = "",
    name3: Dict[str, int] = {"a": 1},
):
    """A function.

    It does something not important but the docstring
    is in ReST style.

    :param name1: A list of bools.
    :type name1: List[bool]
    :param name2: A list or a string.
    This parameter is very important.

    Very important.
        Default is an empty list.
    :type name2: list | str, optional
    :param name3: A dictionary, default is {"a": 1}.
    :type name3: Dict[str, int] optional

    :return result: the result of the function
    :rtype result: int
    """  # noqa: D401
    if name1[0] and len(name2) == 0 and name3["a"] == 1:
        return 1
    return 0


def myfunc_rest2(a: int = 2):
    """Do nothing, a function without indices in docstring.

    :param a:
    """
    return a


def myfunc_numpy(  # pylint: disable=W0102
    name1: List[bool],
    name2: list | str = "",
    name3: Dict[str, int] = {"a": 1},
):
    """A function.

    It does something not important but the docstring
    is in Numpy style.

    Parameters
    ----------
    name1 : List[bool]
        A list of bools.
    name2 : list | str, optional
        A list or a string.
        This parameter is very important.

        Very important.
            Default is an empty list.
    name3 : Dict[str, int] optional
        A dictionary, default is {"a": 1}.

    Returns
    -------
    result : int
        the result of the function
    """  # noqa: D401
    if name1[0] and len(name2) == 0 and name3["a"] == 1:
        return 1
    return 0


def myfunc_numpy2(a: int = 2):
    """Do nothing, a function without indices in docstring.

    Parameters
    ----------
    a
    """
    return a


def myfunc_google(  # pylint: disable=W0102
    name1: List[bool], name2: list | str = "", name3: Dict[str, int] = {"a": 1}
):
    """Do nothing, a function without indices in docstring.

    Args:
        name1 (List[bool]): A list of bools.
        name2 (list | str, optional): A list or a string.
            This parameter is very important.

            Very important.
                Default is an empty list.
        name3 (Dict[str, int], optional): A dictionary, default is {"a": 1}.

    Returns:
        result (int): the result of the function
    """  # noqa: D406 D407
    if name1[0] and len(name2) == 0 and name3["a"] == 1:
        return 1
    return 0


def myfunc_google2(a: int = 2):
    """Do nothing, a function without indices in docstring.

    Args:
        a
    """  # noqa: D406 D407
    return a



"""Test entry points of docstripy."""

import os
import shutil

import pytest
import pytest_check as check

from docstripy.main import write_file, write_files_recursive


def test_main() -> None:
    """Test main entry point."""
    if os.path.exists("tests/tmp"):
        shutil.rmtree("tests/tmp")
    run = os.system("docstripy --help")
    check.equal(run, 0)
    run = os.system(
        "docstripy tests/files -o tests/tmp/google -s google --len 88 --indent 2"
    )
    check.equal(run, 0)
    run = os.system(
        "docstripy tests/files -o tests/tmp/rest -s rest --len 88 --indent 2"
    )
    check.equal(run, 0)
    run = os.system(
        "docstripy tests/files -o tests/tmp/numpy -s numpy --len 88 --indent 2"
    )
    check.equal(run, 0)
    for style in ("google", "numpy", "rest"):
        path = f"tests/tmp/{style}"
        check.is_true(os.path.exists(path), f"Directory not found: {path}")
        for i in range(1, 5):
            path = f"tests/tmp/{style}/test{i}.py"
            check.is_true(os.path.exists(path), f"File not found: {path}")
    with open("tests/tmp/google/empty.py", encoding="utf-8") as file:
        lines = file.readlines()
    check.equal(lines[8], '    """Clean trailing spaces."""\n')
    if os.path.exists("tests/tmp"):
        shutil.rmtree("tests/tmp")


def test_errors(capfd: pytest.CaptureFixture) -> None:
    """Test entry point when there are files errors."""
    with pytest.raises(ValueError, match="Error found during docstring parsing."):
        write_file(
            "tests/wrong_files/file1.py",
            "tests/tmp/file1.py",
            overwrite=False,
            docstr_config={"style": "numpy", "max_len": 88, "indent": 2},
        )
    with pytest.raises(
        ValueError, match="Error found at lines 3-8 during docstring building.*"
    ):
        write_file(
            "tests/wrong_files/file2.py",
            "tests/tmp/file2.py",
            overwrite=False,
            docstr_config={"style": "numpy", "max_len": 88, "indent": 2},
        )
    # Case multiple files
    write_files_recursive(
        "tests/wrong_files",
        "tests/tmp",
        overwrite=False,
        docstr_config={"style": "numpy", "max_len": 88, "indent": 2},
    )
    out, _ = capfd.readouterr()
    check.is_true(out.startswith("Error when parsing file(s):\n"))



"""Test docstring section parser."""

import pytest_check as check

from docstripy.parse_doc.section_ranges import parse_sections_ranges


def test_parse_section_ranges() -> None:
    """Test docstring section parser."""
    with open("tests/files/test2.py", "r", encoding="utf-8") as file:
        lines = file.readlines()
    sections_rest = lines[8:15]
    ranges, style = parse_sections_ranges(sections_rest)
    check.equal(style, "rest")
    expected_ranges = {
        "_title": [0, 2],
        "_parameters": [2, 4],
        "_returns": [4, 6],
    }
    check.equal(ranges, expected_ranges)

    sections_google = lines[23:41]
    ranges, style = parse_sections_ranges(sections_google)
    check.equal(style, "google")
    expected_ranges = {
        "_title": [0, 2],
        "_parameters": [2, 5],
        "_returns": [5, 8],
        "_raises": [15, 17],
        "Examples": [8, 12],
        "Notes": [12, 15],
    }
    check.equal(ranges, expected_ranges)

    sections_numpy = lines[49:70]
    ranges, style = parse_sections_ranges(sections_numpy)
    check.equal(style, "numpy")
    expected_ranges = {
        "_title": [0, 2],
        "_parameters": [2, 7],
        "_returns": [7, 12],
        "Notes": [12, 16],
        "Examples": [16, 21],
    }
    check.equal(ranges, expected_ranges)

    sections_nothing = lines[78:79]
    ranges, style = parse_sections_ranges(sections_nothing)
    check.equal(style, "numpy")  # default
    check.equal(ranges, {"_title": [0, 1]})



"""Docstring building functions for numpy style."""

from typing import Any, Dict, List

from docstripy.line_break import line_break
from docstripy.lines_routines import clean_trailing_empty, clean_trailing_spaces


def build_doc_numpy(
    current_docstring: List[str],
    sections_dict: Dict[str, Any],
    max_len: int,
    indent: int,
) -> List[str]:
    """Build docstring for numpy style."""
    docstring = current_docstring.copy()
    for section_name in (
        "_parameters",
        "_raises",
        "_returns",
        "_yields",
        "_attributes",
    ):
        if section_name in sections_dict and sections_dict[section_name]:
            docstr_sec = []
            header = section_name[1:].capitalize()
            docstr_sec.append("\n")
            docstr_sec.append(header + "\n")
            docstr_sec.append("-" * len(header) + "\n")
            doc_params = build_section_params_numpy(
                sections_dict[section_name],
                max_len=max_len,
                indent=indent,
            )
            if clean_trailing_empty(doc_params):
                docstr_sec.extend(doc_params)
                docstring.extend(docstr_sec)
            else:
                # Error if empty sections
                raise ValueError(f"Empty section found (section {section_name[1:]}).")
    for section_name in sections_dict:
        if not section_name.startswith("_"):
            docstring.append("\n")
            docstring.append(section_name + "\n")
            docstring.append("-" * len(section_name) + "\n")
            docstring.extend(
                line_break(
                    sections_dict[section_name],
                    max_line_length=max_len,
                )
            )
    docstring.append('"""\n')
    docstring = clean_trailing_spaces(docstring)
    return docstring


def build_section_params_numpy(
    param_dicts: List[dict],
    max_len: int,
    indent: int,
) -> List[str]:
    """Build parameters, returns, raises, and attributes sections for numpy style."""
    docstring = []
    for param_dict in param_dicts:
        param_docstring = []
        first_line = ""
        if "name" in param_dict and param_dict["name"]:
            first_line += param_dict["name"] + " : "
        if "type" in param_dict and param_dict["type"]:
            first_line += param_dict["type"]
            if "optional" in param_dict and param_dict["optional"]:
                first_line += ", optional"
        if "description" in param_dict and param_dict["description"]:
            param_docstring.extend(param_dict["description"])
        if "default" in param_dict and param_dict["default"]:
            if (
                not param_docstring
                or "description" not in param_dict
                or len(param_docstring[-1]) > max_len * 0.75
            ):
                line = "By default, " + param_dict["default"] + ".\n"
                param_docstring.append(line)
            else:
                param_docstring[-1] = param_docstring[-1].rstrip("\n")
                param_docstring[-1] += " By default, " + param_dict["default"] + ".\n"
        param_docstring = [first_line + "\n"] + line_break(
            param_docstring, max_len - indent
        )
        for i, line in enumerate(param_docstring):
            if i > 0:
                param_docstring[i] = " " * indent + line
        docstring.extend(param_docstring)
    return docstring



"""Tests for file parsing functions."""

import pytest_check as check

from docstripy.parse_doc.main_parser import parse_docstring


def test_parse_docstring() -> None:
    """Test docstring file parser."""
    with open("tests/files/test1.py", encoding="utf-8") as file:
        lines_test1 = file.readlines()
    range_docstr, _, to_insert = parse_docstring(lines_test1)
    check.equal(range_docstr, [[8, 24], [32, 33], [43, 47], [51, 52]])
    check.equal(to_insert, [False, False, False, True])
    with open("tests/files/test4.py", encoding="utf-8") as file:
        lines_test4 = file.readlines()
    _, sections_list, _ = parse_docstring(lines_test4)
    check.equal(len(sections_list), 3)
    expected_dict = {
        "_escaped": True,
        "_title": [
            "A function.\n",
            "\n",
            "It does something not important but the title\n",
            "is in multiple line.\n",
        ],
        "_parameters": [
            {
                "name": "name1",
                "type": "List[bool]",
                "optional": False,
                "description": ["A list of bools.\n"],
                "default": "",
            },
            {
                "name": "name2",
                "type": "list | str",
                "optional": True,
                "description": [
                    "A list or a string.\n",
                    "\n",
                    "This parameter is very important.\n",
                ],
                "default": '""',
            },
            {
                "name": "name3",
                "type": "Dict[str, int]",
                "optional": True,
                "description": ["A dictionary.\n"],
                "default": '{"a,b": 1}',
            },
        ],
        "_raises": [
            {
                "name": "ValueError",
                "type": "",
                "description": ["When something happened.\n"],
            },
            {
                "name": "ValueError",
                "type": "",
                "description": ["When something else happened.\n"],
            },
        ],
        "_returns": [
            {
                "name": "result",
                "type": "int",
                "description": ["The first result of the function.\n"],
            },
            {
                "name": "",
                "type": "list",
                "description": [],
            },
            {
                "name": "result3",
                "type": "int",
                "description": [],
            },
        ],
        "_yields": [
            {
                "name": "",
                "type": "int",
                "description": ["A number.\n"],
            },
        ],
        "_attributes": [{"name": "nothing", "type": "", "description": []}],
        "Example": [
            ">>> myfunc_rest([True])\n",
            "1, [], 1\n",
            """>>> myfunc_rest([True], "\\n")\n""",
            "0, [], 0\n",
        ],
        "Notes": [
            "* This always works\n",
            "* This is a note\n",
        ],
    }
    for i in range(3):
        for key, val in expected_dict.items():
            if key != "_parameters":
                check.equal(
                    sections_list[i][key],
                    val,
                    f"Error with style {i} and key {key}.",
                )
        for j, param in enumerate(sections_list[i]["_parameters"]):
            check.equal(
                param,
                expected_dict["_parameters"][j],  # type: ignore
                f"Error with style {i} and parameter {j}.",
            )



# type: ignore
"""A test module.

Contains some functions for testing the parser.
"""


def factorial_rest(n):
    r"""Return factorial of '\n'.

    :param n: Number to compute factorial of.
    :type n: int
    :return: Factorial of n.
    :rtype: int
    """
    if n < 0:
        raise ValueError("Cannot compute factorial of negative number.")
    if n == 0:
        return 1
    return n * factorial_rest(n - 1)


def factorial_google(n):
    r"""Return factorial of '\n'.

    Args:
        n (int): Number to compute factorial of.

    Returns:
        int: Factorial of n.

    Examples:
    >>> factorial_google(5)
    120

    Notes:
        This is a note.

    Raises:
        ValueError: If n is negative.
    """  # noqa: D406, D407, D205
    if n < 0:
        raise ValueError("Cannot compute factorial of negative number.")
    if n == 0:
        return 1
    return n * factorial_google(n - 1)


def factorial_numpy(n):
    r"""Return factorial of '\n'.

    Parameters
    ----------
    n : int
        Number to compute factorial of.

    Returns
    -------
    int :
        Factorial of n.

    Notes
    -----
    This is a note.

    Examples
    --------
    >>> factorial_numpy(5)
    120
    """
    if n < 0:
        raise ValueError("Cannot compute factorial of negative number.")
    if n == 0:
        return 1
    return n * factorial_google(n - 1)


def factorial_nothing(n):
    r"""Return factorial of '\n'."""
    if n < 0:
        raise ValueError("Cannot compute factorial of negative number.")
    if n == 0:
        return 1
    return n * factorial_google(n - 1)



"""Classes and functions to manage difference between files."""

from typing import List, Tuple, Union


def satenize_ranges(ranges: List[List[int]]) -> List[List[int]]:
    """Check if the ranges are overlapping."""
    for i, range1 in enumerate(ranges):
        for range2 in ranges[i + 1 :]:
            if range1[0] <= range2[0] < range1[1]:
                raise ValueError("Found overlapping ranges.")
            if range2[0] <= range1[0] < range2[1]:
                raise ValueError("Found overlapping ranges.")
    return ranges


def order_lists(
    ranges: List[List[int]],
    lines: List[str],
    to_insert: List[bool],
) -> Tuple[List[List[int]], List[str], List[bool]]:
    """Order the ranges in descending order and lines accordingly."""
    sort_indices = sorted(range(len(ranges)), key=lambda i: ranges[i][0])
    ranges = [ranges[i] for i in sort_indices[::-1]]
    lines = [lines[i] for i in sort_indices[::-1]]
    to_insert = [to_insert[i] for i in sort_indices[::-1]]
    return ranges, lines, to_insert


def split_line(line: str) -> List[str]:
    r"""Split a line in multiple lines using '\n'."""
    if line[-1] == "\n":
        line = line[:-1]
    return [single_line + "\n" for single_line in line.split("\n")]


def apply_diff(
    ranges: List[List[int]],
    lines: List[str],
    old_lines: List[str],
    *,
    to_insert: Union[bool, List[bool]] = False,
) -> List[str]:
    """Apply the difference to a list of lines.

    Parameters
    ----------
    ranges : List[List[int]]
        List of ranges of docstring line numbers [start, end]. Includes line at start,
        excludes line at end.
    lines : List[str]
        List of lines to add to the file.
    old_lines : List[str]
        List of lines of the original file.
    to_insert : Union[bool, List[bool]], optional
        Whether to insert the lines or overwrite them instead.
        If a list is given, it should have the same
        length as ranges. Otherwise, the same value will be used for all ranges.
        By default False.
    """
    if isinstance(to_insert, bool):
        to_insert_list = [to_insert] * len(ranges)
    else:
        to_insert_list = to_insert
    ranges, lines, to_insert_list = order_lists(ranges, lines, to_insert_list)
    ranges = satenize_ranges(ranges)
    new_lines = old_lines.copy()
    for i, line in enumerate(lines):
        start, end = ranges[i]
        to_insert_line = to_insert_list[i]
        if not to_insert_line:
            del new_lines[start:end]
        new_lines[start:start] = split_line(line)
    return new_lines



"""Routines on code source lines."""

from typing import List, Tuple


def add_eol(lines: List[str]) -> List[str]:
    """Add end of lines."""
    new_lines = lines.copy()
    for i, line in enumerate(lines):
        if not line.endswith("\n"):
            new_lines[i] = line + "\n"
    return new_lines


def remove_indent(lines: List[str]) -> List[str]:
    """Remove indent from lines."""
    indent = find_indent(lines)
    new_lines = lines.copy()
    for i, line in enumerate(lines):
        new_lines[i] = line[indent:] if line not in ("", "\n") else line
    return new_lines


def add_indent(lines: List[str], indent: int) -> List[str]:
    """Add indent from lines."""
    new_lines = lines.copy()
    for i, line in enumerate(lines):
        new_lines[i] = " " * indent + line if line not in ("", "\n") else line
    return new_lines


def find_indent(lines: List[str]) -> int:
    """Find indentation of a docstring."""
    if len([line for line in lines if line not in ("", "\n")]) == 0:
        return 0
    return min(
        len(line) - len(line.lstrip()) for line in lines if line not in ("", "\n")
    )


def find_prefix(
    lines: List[str],
    prefix_start: Tuple[str, ...],
    prefix_continue: Tuple[str, ...],
    *,
    dash: bool = False,
) -> Tuple[int, int]:
    """Find the index of consecutive lines starting with prefix.

    Parameters
    ----------
    lines : List[str]
        Lines of the docstring.
    prefix_start : Tuple[str, ...]
        Prefixes string of the first line.
    prefix_continue : Tuple[str, ...]
        Prefixes string of the following lines.
    dash : bool, optional
        Whether to expect a dash line after the first line, by default False.

    Returns
    -------
    start : int
        Index of the first line (or -1 if not found).
    end : int
        Index of the last line (or -1 if not found).
    """
    start, end = -1, -1
    if dash:
        prefix_continue += ("---",)
    for i, line in enumerate(lines):
        if line.startswith(prefix_start):
            if dash and len(lines) <= i + 1:
                continue
            if dash and len(lines) > i + 1 and not lines[i + 1].startswith("---"):
                continue
            start = i
            for j, line in enumerate(lines[i + 1 :]):
                if not line.startswith(prefix_continue):
                    end = i + j + 1
                    break
            break
    return start, end


def clean_leading_empty(lines: List[str]) -> List[str]:
    """Clean leading empty lines."""
    new_lines = lines.copy()
    while len(new_lines) > 0 and new_lines[0].strip(" ") == "\n":
        del new_lines[0]
    return new_lines


def clean_trailing_empty(lines: List[str]) -> List[str]:
    """Clean trailing empty lines."""
    new_lines = lines.copy()
    while len(new_lines) > 0 and new_lines[-1].strip(" ") == "\n":
        del new_lines[-1]
    return new_lines


def remove_quotes(lines: List[str]) -> Tuple[List[str], bool]:
    """Remove triple quotes and return whether the docstring is escaped or not."""
    escaped = False
    new_lines = lines.copy()
    for comment_pattern in ("'''#", "''' #", "'''  #", '"""#', '""" #', '"""  #'):
        if comment_pattern in new_lines[-1]:
            ind = new_lines[-1].index(comment_pattern)
            new_lines[-1] = new_lines[-1][:ind]
            break

    for i, line in enumerate(lines):
        if "r'''" in line or 'r"""' in line:
            escaped = True
            new_lines[i] = line.replace("r'''", "").replace('r"""', "")
        if "'''" in new_lines[i] or '"""' in new_lines[i]:
            new_lines[i] = line.replace("'''", "").replace('"""', "")
    new_lines = add_eol(new_lines)  # Add end of lines if removed
    return new_lines, escaped


def clean_trailing_spaces(lines: List[str]) -> List[str]:
    """Clean trailing spaces."""
    new_lines = []
    for line in lines:
        have_eol = line and line[-1] == "\n"
        line = line.rstrip("\n")
        line = line.rstrip(" ")
        if have_eol:
            line += "\n"
        new_lines.append(line)
    return new_lines


def clean_comment(lines: List[str]) -> List[str]:
    """Clean comment lines."""
    new_lines = []
    for line in lines:
        if "#" in line:
            line = line[: line.index("#")]
        new_lines.append(line)
    return new_lines



"""File parsing functions."""

from typing import List, Tuple


def parse_ranges(lines: List[str]) -> Tuple[List[List[int]], List[List[int]]]:
    """Parse source code lines ranges.

    Parameters
    ----------
    lines : List[str]
        List of lines to parse.

    Returns
    -------
    ranges_docstr : List[List[int]]
        Ranges of lines containing docstrings.
    ranges_def : List[List[int]]
        Ranges of lines containing function definitions.
    """
    ranges_docstr: List[List[int]] = []
    ranges_def: List[List[int]] = []
    current_docstr_range: List[int] = []
    current_def_range: List[int] = []
    in_docstr, in_def = False, False
    for ind_line, line in enumerate(lines):
        ranges_docstr, current_docstr_range, in_docstr = docstring_parse_range(
            ranges=ranges_docstr,
            current_range=current_docstr_range,
            in_docstr=in_docstr,
            line=line,
            ind_line=ind_line,
        )
        if not in_docstr:
            ranges_def, current_def_range, in_def = def_parse_range(
                ranges=ranges_def,
                current_range=current_def_range,
                in_def=in_def,
                line=line,
                ind_line=ind_line,
            )
    return ranges_docstr, ranges_def


def docstring_parse_range(
    ranges: List[List[int]],
    current_range: List[int],
    *,
    in_docstr: bool,
    line: str,
    ind_line: int,
) -> Tuple[List[List[int]], List[int], bool]:
    """Parse docstring lines range from source code."""
    docstring_starters = ('"""', "'''", 'r"""', "r'''")
    strip_line = line.strip()
    if strip_line.startswith(docstring_starters) and line.startswith(" "):
        # Case one-line docstring
        if (
            strip_line.count('"""') > 1 or strip_line.count("'''") > 1
        ) and not in_docstr:
            ranges.append([ind_line, ind_line + 1])
        else:
            # Case multi-line docstring
            if not in_docstr:
                current_range = [ind_line]
                in_docstr = True
            else:
                current_range.append(ind_line + 1)
                ranges.append(current_range)
                in_docstr = False
    elif in_docstr and strip_line.endswith('"""') or strip_line.endswith("'''"):
        current_range.append(ind_line + 1)
        ranges.append(current_range)
        in_docstr = False
    return ranges, current_range, in_docstr


def def_parse_range(
    ranges: List[List[int]],
    current_range: List[int],
    *,
    in_def: bool,
    line: str,
    ind_line: int,
) -> Tuple[List[List[int]], List[int], bool]:
    """Parse docstring lines range from source code."""
    strip_line = line.rsplit("#", maxsplit=1)[0].strip()
    if not in_def and strip_line.startswith("def "):
        if strip_line.endswith(":"):
            # Case one-line def
            ranges.append([ind_line, ind_line + 1])
        else:
            # Case multi-line def
            current_range = [ind_line]
            in_def = True
    elif in_def and strip_line.endswith(":"):
        current_range.append(ind_line + 1)
        ranges.append(current_range)
        in_def = False
    return ranges, current_range, in_def



"""Parameter section parsing functions."""

from functools import partial
from typing import List, Tuple

from docstripy.google.parse_doc import parse_params as parse_params_google
from docstripy.lines_routines import add_eol, clean_trailing_empty
from docstripy.numpy.parse_doc import parse_params as parse_params_numpy
from docstripy.rest.parse_doc import parse_params as parse_params_rest


def parse_params_all(
    lines: List[str],
    style: str,
    section_name: str = "param",
) -> List[dict]:
    r"""Parse parameters, raises, returns, yields and attributes section.

    Parameters
    ----------
    lines : List[str]
        Lines of the docstring containing parameters descriptions.
    style : str
        Format style of the docstring (one of 'numpy', 'google' or 'rest').
    section_name : str
        Name of the section to parse. One of "param", "return", "yield", "attribute",
        "raises".
        By default, "param".

    Returns
    -------
    params_dict : dict
        Parameters dictionary with the following keys:

        type : dict
            Dictionary of parameter names and their types.
        optional : dict
            Dictionary of parameter names and whether they are optional.
        description : dict
            Dictionary of parameter names and their descriptions.
        default : dict
            Dictionary of parameter names and their default values.

    Example
    -------
    One can parse a docstring as follows:
    [
        {
            'name': 'param1',
            'type': 'int',
            'optional': False,
            'description': ['Description of param1\n'],
        },
        {
            'name': 'param2',
            'type': 'str',
            'optional': True,
            'description': ['Description of param2\n'],
            'default': '""',
        },
    ],
    """
    pattern_type_rest = "rtype" if section_name in ("return", "yield") else "type"
    parse_params_rest_patterns = partial(
        parse_params_rest,
        pattern_type=pattern_type_rest,
    )
    parse_funcs = {
        "rest": parse_params_rest_patterns,
        "google": parse_params_google,
        "numpy": parse_params_numpy,
    }
    parse_func = parse_funcs[style]
    params_list = parse_func(lines=lines, section_name=section_name)  # type: ignore

    for param_dict in params_list:
        description = param_dict["description"]
        if section_name == "param":
            description, default_val = extract_default_value(param_dict["description"])
            param_dict["default"] = postprocess_default_val(default_val)
            if param_dict["default"] != "":
                param_dict["optional"] = True
        param_dict["description"] = postprocess_description(description)

    if section_name != "param":
        # Remove optional info
        for param_dict in params_list:
            del param_dict["optional"]

    return params_list


def extract_default_value(lines: List[str]) -> Tuple[List[str], str]:
    """Extract default value from description."""
    # Lines merging
    line = "".join(lines)

    patterns = [
        "default is ",
        "Default is ",
        "defaults to ",
        "default to ",
        "Defaults to ",
        "Default to ",
        "by default ",
        "By default ",
        "defaults : ",
        "Defaults : ",
        "default : ",
        "Default : ",
        "defaults: ",
        "Defaults: ",
        "default: ",
        "Default: ",
    ]
    patterns_comma1 = [", " + pattern for pattern in patterns]
    patterns_comma2 = [pattern + ", " for pattern in patterns]
    patterns = patterns_comma2 + patterns_comma1 + patterns
    default = ""
    for pattern in patterns:
        if pattern in line:
            default_split1, default_split_2 = line.rsplit(pattern, maxsplit=1)
            default = default_split_2.split("\n")[0].split(".")[0].strip()
            if default + "." in default_split_2:
                line = default_split1 + default_split_2.replace(default + ".", "")
            else:
                line = default_split1 + default_split_2.replace(default, "")
            break
    # Reverse lines merging
    lines = line.split("\n")
    lines = [line + "\n" for line in lines]
    return lines, default


def postprocess_description(lines: List[str]) -> List[str]:
    """Post process description."""
    lines = add_eol(lines)
    lines = clean_trailing_empty(lines)
    if lines and len(lines[-1]) > 1 and lines[-1][-2] not in (".", "?", "!"):
        lines[-1] = lines[-1][:-1] + ".\n"
    # Capitalize first letter
    if len(lines) > 0:
        first_line_desc = lines[0]
        if 97 <= ord(first_line_desc[0]) <= 122:
            cap_letter = first_line_desc[0].capitalize()
            lines[0] = cap_letter + first_line_desc[1:]
    return lines


def postprocess_default_val(default: str) -> str:
    """Post process default value string."""
    if default == "":
        return default
    default.replace("\n", " ")  # remove new lines characters
    if default.endswith("."):
        default = default[:-1]
    return default
